from diffusion_webui.controlnet.controlnet_canny import (
    stable_diffusion_controlnet_canny,
    stable_diffusion_controlnet_canny_app,
)
from diffusion_webui.controlnet.controlnet_depth import (
    stable_diffusion_controlnet_depth,
    stable_diffusion_controlnet_depth_app,
)
from diffusion_webui.controlnet.controlnet_hed import (
    stable_diffusion_controlnet_hed,
    stable_diffusion_controlnet_hed_app,
)
from diffusion_webui.controlnet.controlnet_mlsd import (
    stable_diffusion_controlnet_mlsd,
    stable_diffusion_controlnet_mlsd_app,
)
from diffusion_webui.controlnet.controlnet_pose import (
    stable_diffusion_controlnet_pose,
    stable_diffusion_controlnet_pose_app,
)
from diffusion_webui.controlnet.controlnet_scribble import (
    stable_diffusion_controlnet_scribble,
    stable_diffusion_controlnet_scribble_app,
)
from diffusion_webui.controlnet.controlnet_seg import (
    stable_diffusion_controlnet_seg,
    stable_diffusion_controlnet_seg_app,
)
from diffusion_webui.controlnet_inpaint.controlnet_inpaint_app import (
    stable_diffusion_inpiant_controlnet_canny,
    stable_diffusion_inpiant_controlnet_canny_app,
)
from diffusion_webui.stable_diffusion.img2img_app import (
    stable_diffusion_img2img,
    stable_diffusion_img2img_app,
)
from diffusion_webui.stable_diffusion.inpaint_app import (
    stable_diffusion_inpaint,
    stable_diffusion_inpaint_app,
)
from diffusion_webui.stable_diffusion.text2img_app import (
    stable_diffusion_text2img,
    stable_diffusion_text2img_app,
)
